package BuildAux::FileUtils;

=head1 NAME

BuildAux::FileUtils - handling files

=head1 SYNOPSIS

  use BuildAux::FileUtils

=head1 DESCRIPTION

This perl module provides various general purpose file handling functions.

=cut

use strict;
use BuildAux::Verbose;
use File::Basename;

use Exporter;

our @ISA = qw (Exporter);
our @EXPORT = qw (&contents
                  &ensure_dir &ensure_parent_dir
                  &install
                  &same_file);

=item C<contents ($file)>

The contensts of C<$file> as a single string.

=cut

sub contents ($)
{
  my ($file) = @_;
  local $/;                     # Turn on slurp-mode.
  use IO::File;
  my $f = new IO::File "< $file" or die "cannot open $file: $!";
  my $contents = $f->getline or die "cannot read $file: $!";
  $f->close;
  $contents;
}

=item C<ensure_dir($dir)>

Create, if needed, the directory C<$dir>.

=cut

sub ensure_dir ($)
{
  my ($dir) = @_;
  if (! -d $dir)
    {
      use File::Path;
      mkpath ($dir, { verbose => $verbose, })
        or die "$0: cannot create $dir: $!";
    }
}

=item C<ensure_parent_dir($filename)>

Create, if needed, the directory that will contain I<$filename>.

=cut

sub ensure_parent_dir ($)
{
  my ($file) = @_;
  ensure_dir (dirname ($file));
}

=item C<install($from, $to)>

Copy the file C<$from> as the file C<$to>.  Create parent directories
if needed.

Preserve permissions.  This avoids spurious warnings from ldd that
does not like that shared objects are not executable.

=cut

sub install($$)
{
  my ($from, $to) = @_;
  verbose 3, "copy $from $to";
  ensure_parent_dir($to);

  use File::Copy;
  File::Copy::copy $from, $to
    or die "$me: copy $from $to failed: $!\n";

  my $mode = (stat($from))[2] & 0777;
  chmod $mode, $to
    or die "$me: cannot chmod $to failed: $!\n";
}


=item C<same_file($path1, $path1)>

Whether C<$path1> and C<$path2> point to the same file.  We do not
work on path names because there are too many ways to name a single
file, in particular when winepath goes into the scene.

=cut

sub same_file ($$)
{
  my ($path1, $path2) = @_;
  my ($dev1, $ino1) = stat ($path1);
  my ($dev2, $ino2) = stat ($path2);
  $dev1 == $dev2 && $ino1 == $ino2;
}



1; # for require

### Setup "Gostai" style for perl-mode and cperl-mode.
## Local Variables:
## perl-indent-level: 2
## perl-continued-statement-offset: 2
## perl-continued-brace-offset: -2
## perl-brace-offset: 0
## perl-brace-imaginary-offset: 0
## perl-label-offset: -2
## cperl-indent-level: 2
## cperl-brace-offset: 0
## cperl-continued-brace-offset: -2
## cperl-label-offset: -2
## cperl-extra-newline-before-brace: t
## cperl-merge-trailing-else: nil
## cperl-continued-statement-offset: 2
## End:
